/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef CACHE_H
#define CACHE_H

#include <stdio.h>
#include "ss.h"
#include "memory.h"

/* cache implementation uses a hash table lookup to speed block access
   when the CACHE_HIGHLY_ASSOC() returns non-zero; otherwise, the hash
   tables are not allocated, and cache blocks are located with a
   linear search of the MRU block list */
#define CACHE_HIGHLY_ASSOC(cp)	((cp)->assoc > 4)

/* cache replacement policy */
enum cache_policy {
  LRU,		/* way list order: MRU -> LRU */
  Random,	/* way list order: arbitrary */
  FIFO		/* way list order: oldest block -> newest block */
};

struct cache_blk
{
  struct cache_blk *way_next;	/* next block in the MRU chain */
  struct cache_blk *way_prev;	/* previous block in the MRU chain */
  struct cache_blk *hash_next;	/* next block in the hash bucket chain, only
				   used in highly-associative caches */
  /* since hash table lists are typically small, there is no previous
     pointer, deletion requires a trip through the hash table bucket list */
  SS_ADDR_TYPE tag;		/* data block tag value */
  unsigned long status;		/* block status */
  SS_TIME_TYPE ready;		/* when will this block be accessable */
  char *user_data;		/* pointer to user defined data, e.g.,
				   pre-decode data or physical page address */
  /* DATA should be pointer-aligned due to preceeding field */
  char data[0];			/* actual data block starts here, block size
				   should probably be a multiple of 8 */
};

/* block status values */
#define CACHE_BLK_VALID		0x00000001	/* block in valid, in use */
#define CACHE_BLK_DIRTY		0x00000002	/* dirty block */

struct cache_set
{
  struct cache_blk **hash;	/* hash table: for fast access w/assoc, NULL
				   for low-assoc caches */
  struct cache_blk *way_head;	/* head of way list */
  struct cache_blk *way_tail;	/* tail pf way list */
  struct cache_blk *blks;	/* cache blocks, allocated sequentially, so
				   this pointer can also be used for random
				   access to cache blocks */
};

struct cache
{
  /* parameters */
  char *name;			/* cache name */
  int nsets;			/* number of sets */
  int bsize;			/* block size in bytes */
  int balloc;			/* maintain cache contents? */
  int usize;			/* user allocated data size */
  int assoc;			/* cache associativity */
  enum cache_policy policy;	/* cache replacement policy */
  unsigned long hit_latency;	/* cache hit latency */

  /* miss/replacement handlers, read/write BSIZE bytes starting at BADDR
     from/into cache block BLK, returns the latency of the operation
     if initiated at NOW, returned latencies indicate how long it takes
     for the cache access to continue (e.g., fill a write buffer), the
     miss/repl functions are required to track how this operation will
     effect the latency of later operations (e.g., write buffer fills),
     if !BALLOC, then just return the latency; BLK_ACCESS_FN is also
     responsible for generating any user data and incorporating the latency
     of that operation */
  unsigned long (*blk_access_fn)(enum mem_cmd cmd,
				 SS_ADDR_TYPE baddr, int bsize,
				 struct cache_blk *blk, SS_TIME_TYPE now);

  /* derived data, for fast decoding */
  int hsize;			/* cache set hash table size */
  SS_ADDR_TYPE blk_mask;
  int set_shift;
  SS_ADDR_TYPE set_mask;	/* use *after* shift */
  int tag_shift;
  SS_ADDR_TYPE tag_mask;	/* use *after* shift */
  SS_ADDR_TYPE tagset_mask;	/* used for fast hit detection */

  /* bus resource */
  SS_TIME_TYPE bus_free;	/* time when bus to next level of cache is
				   free, NOTE: the bus model assumes only a
				   single, fully-pipelined port to the next
 				   level of memory that requires the bus only
 				   one cycle for cache line transfer (the
 				   latency of the access to the lower level
 				   may be more than one cycle, as specified
 				   by the miss handler */
 

  /* stats */
  SS_COUNTER_TYPE hits;
  SS_COUNTER_TYPE misses;
  SS_COUNTER_TYPE replacements;
  SS_COUNTER_TYPE invalidations;

  /* last block to hit */
  SS_ADDR_TYPE last_tagset;
  struct cache_blk *last_blk;

  /* data blocks */
  char *data;			/* pointer to data blocks allocation */
  struct cache_set sets[0];	/* each entry is a set */
};

/* create and initialize a general cache structure, see "struct cache"
   definition for a description of parameters */
struct cache *
cache_create(char *name,
	     int nsets, int bsize, int balloc, int usize,
	     int assoc, enum cache_policy policy,
	     unsigned long (*blk_access_fn)(enum mem_cmd cmd,
					    SS_ADDR_TYPE baddr, int bsize,
					    struct cache_blk *blk,
					    SS_TIME_TYPE now),
	     unsigned long hit_latency);

/* parse policy */
enum cache_policy cache_char2policy(char c);

/* print cache configuration */
void cache_config(struct cache *cp, FILE *stream);

/* print cache stats */
void cache_stats(struct cache *cp, FILE *stream);

/* access a cache, perform a CMD operation on cache CP at address ADDR,
   places NBYTES of data at *P, returns latency of operation if initiated
   at NOW, places pointer to block user data in *UDATA, *P is untouched if
   cache blocks are not allocated (!CP->BALLOC), UDATA should be NULL if no
   user data is attached to blocks */
unsigned long
cache_access(struct cache *cp, enum mem_cmd cmd, SS_ADDR_TYPE addr,
	     void *p, int nbytes, SS_TIME_TYPE now, char **udata,
	     SS_ADDR_TYPE *repl_addr);
 
/* return non-zero if block containing address ADDR is contained in cache
   CP, this interface is used primarily for debugging and asserting cache
   invariants */
int
cache_probe(struct cache *cp, SS_ADDR_TYPE addr);

/* cache access functions, these are safe, they check alignment and
   permissions */
#define cache_double(cp, cmd, addr, p, now, udata)	\
  cache_access(cp, cmd, addr, p, sizeof(double), now, udata)
#define cache_float(cp, cmd, addr, p, now, udata)	\
  cache_access(cp, cmd, addr, p, sizeof(float), now, udata)
#define cache_dword(cp, cmd, addr, p, now, udata)	\
  cache_access(cp, cmd, addr, p, sizeof(long long), now, udata)
#define cache_word(cp, cmd, addr, p, now, udata)	\
  cache_access(cp, cmd, addr, p, sizeof(long), now, udata)
#define cache_half(cp, cmd, addr, p, now, udata)	\
  cache_access(cp, cmd, addr, p, sizeof(short), now, udata)
#define cache_byte(cp, cmd, addr, p, now, udata)	\
  cache_access(cp, cmd, addr, p, sizeof(char), now, udata)

/* flush the entire cache; returns latency of the operation */
unsigned long cache_flush(struct cache *cp, SS_TIME_TYPE now);

/* flush the block containing ADDR from the cache CP; returns the latency of
   the operation */
unsigned long
cache_flush_addr(struct cache *cp, SS_ADDR_TYPE addr, SS_TIME_TYPE now);

#endif /* CACHE_H */
