/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef MEMORY_H
#define MEMORY_H

#include <stdio.h>
#include "endian.h"
#include "ss.h"

enum mem_cmd { Read, Write };

typedef void
(*mem_access_fn)(enum mem_cmd cmd,	/* Read or Write */
		 unsigned long addr,	/* address to access */
		 void *p,		/* where to copy to/from */
		 int nbytes);		/* xfer length */

/* memory indirect table size (upper mem is not used) */
#define MEM_TABLE_SIZE		0x8000 /* was: 0x7fff */

/* memory block size, in bytes */
#define MEM_BLOCK_SIZE		0x10000

/* memory access macros */
#if 0
#define MEM_BLOCK(addr) 	(((SS_ADDR_TYPE)(addr)) >> 16)
#else
#define MEM_BLOCK(addr) 	((((SS_ADDR_TYPE)(addr)) >> 16) & 0xffff)
#endif
#if 0
#define MEM_OFFSET(addr)	((((SS_ADDR_TYPE)(addr)) << 16) >> 16)
#else
#define MEM_OFFSET(addr)	((addr) & 0xffff)
#endif

/* top of the data segment */
extern SS_ADDR_TYPE mem_brk_point;

/* lowest address accessed on the stack */
extern SS_ADDR_TYPE mem_stack_min;

extern char *mem_table[MEM_TABLE_SIZE];

#define __UNCHK_MEM_ACCESS(type, addr)					\
  (*((type *)(mem_table[MEM_BLOCK(addr)] + MEM_OFFSET(addr))))

#define __MEM_ACCESS(type, addr)					\
  ((!mem_table[MEM_BLOCK(addr)]						\
    ? (mem_table[MEM_BLOCK(addr)] = mem_newblock())			\
    : 0),								\
   *((type *)(mem_table[MEM_BLOCK(addr)] + MEM_OFFSET(addr))))

/* fast memory access macros, these are unsafe, use lower case versions
   to enable alignment and permission checks; note, these macros may by used
   as l-value or r-values; note, all returned integer values are unsigned */
#define MEM_WORD(addr)		__MEM_ACCESS(unsigned long, addr)
#define MEM_HALF(addr)		__MEM_ACCESS(unsigned short, addr)
#define MEM_BYTE(addr)		__MEM_ACCESS(unsigned char, addr)

/* memory access functions, these are safe, swapped, and check alignment
   and permissions */
#define mem_read_word(cmd, addr, p)					\
  (mem_access(cmd, addr, p, sizeof(unsigned long)),			\
   *((unsigned long *)p) = SWAP_WORD(*((unsigned long *)p)))
#define mem_write_word(cmd, addr, p)					\
  ({ unsigned long _temp = SWAP_WORD(*((unsigned long *)p));		\
     mem_access(cmd, addr, &_temp, sizeof(unsigned long)); })

#define mem_read_half(cmd, addr, p)					\
  (mem_access(cmd, addr, p, sizeof(unsigned short)),			\
   *((unsigned short *)p) = SWAP_HALF(*((unsigned short *)p)))
#define mem_write_half(cmd, addr, p)					\
  ({ unsigned short _temp = SWAP_HALF(*((unsigned short *)p));		\
     mem_access(cmd, addr, &_temp, sizeof(unsigned short)); })

#define mem_read_byte(cmd, addr, p)					\
  mem_access(cmd, addr, p, sizeof(char))
#define mem_write_byte(cmd, addr, p)					\
  mem_access(cmd, addr, p, sizeof(char))


void mem_access(enum mem_cmd cmd, SS_ADDR_TYPE addr, void *p, int nbytes);

char *mem_newblock(void);

int
mem_strcpy(mem_access_fn mem_fn, enum mem_cmd cmd, SS_ADDR_TYPE addr, char *s);

void mem_bcopy(mem_access_fn mem_fn, enum mem_cmd cmd, SS_ADDR_TYPE addr,
	       void *p, int nbytes);

void mem_bcopy4(mem_access_fn mem_fn, enum mem_cmd cmd, SS_ADDR_TYPE addr,
		void *p, int nbytes);

void mem_bzero(mem_access_fn mem_fn, SS_ADDR_TYPE addr, int nbytes);

/* memory system-specific options */
extern char *mem_optstring;

/* initialize memory system */
void mem_init(void);

/* parse memory system-specific options */
void mem_options(int argc, char **argv);

/* print out memory system configuration */
void mem_config(FILE *stream);

/* dump memory system stats */
void mem_stats(FILE *stream);

/* dump a block of memory */
void mem_dump(mem_access_fn mem_fn, SS_ADDR_TYPE addr, int len, FILE *stream);

#endif /* MEMORY_H */
