/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"
#include "cache.h"
#include "loader.h"
#include "syscall.h"
#include "sim.h"

/* track number of insn and refs */
static SS_COUNTER_TYPE num_insn = 0;
static SS_COUNTER_TYPE num_refs = 0;

static struct cache *icache = NULL;
static struct cache *dcache = NULL;
static struct cache *tlb = NULL;

static int flush_on_syscalls = FALSE;

static unsigned long
blk_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE baddr, int bsize,
	      struct cache_blk *blk, SS_TIME_TYPE now)
{
  mem_access(cmd, baddr, &blk->data[0], bsize);
  return 0;
}

static unsigned long
tlb_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE baddr, int bsize,
	      struct cache_blk *blk, SS_TIME_TYPE now)
{
  /* no real memory access */
  assert(blk->user_data);
  /* fake translation */
  *((SS_ADDR_TYPE *)(blk->user_data)) = 0;
  return 0;
}

/* simulator-specific options */
char *sim_optstring = "fd:i:t:";

void
sim_options(int argc, char **argv)
{
  char c;

  /* parse options */
  getopt_init();
  while ((c = getopt_next(argc, argv, sim_optstring)) != EOF)
    {
      switch (c) {
      case 'i':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a I-cache */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad I-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
	  icache =
	    cache_create(name, nsets, bsize, /* balloc */TRUE, /* usize */0,
			 assoc, cache_char2policy(c), blk_access_fn,
			 /* hit latency */0);
	}
	break;
      case 'd':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a D-cache */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad D-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
	  dcache =
	    cache_create(name, nsets, bsize, /* balloc */TRUE, /* usize */0,
			 assoc, cache_char2policy(c), blk_access_fn,
			 /* hit latency */0);
	}
	break;
      case 't':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a TLB */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad TLB parms: <name>:<nsets>:<page_size>:<assoc>:<repl>");
	  tlb =
	    cache_create(name, nsets, bsize, /* balloc */FALSE,
			 /* usize */sizeof(SS_ADDR_TYPE), assoc,
			 cache_char2policy(c), tlb_access_fn,
			 /* hit latency */0);
	}
	break;
      case 'f':
	flush_on_syscalls = TRUE;
	break;
      }
    }
}

void
sim_config(FILE *stream)
{
  if (icache)
    cache_config(icache, stream);
  if (dcache)
    cache_config(dcache, stream);
  if (flush_on_syscalls)
    info("flushing D-cache on system calls...");
  else
    info("running system calls through warm D-cache...");
  if (tlb)
    cache_config(tlb, stream);
}

/* set to non-zero when simulator should dump statistics */
int sim_dump_stats = FALSE;

void
sim_stats(FILE *stream)
{
  extern time_t start_time;
  time_t elapsed = MAX(time((time_t *)NULL) - start_time, 1);

  info("sim: executed %.0f instructions", (double)num_insn);
  info("sim: accessed memory %.0f times", (double)num_refs);
  info("sim: simulation time: %s (%f insts/sec)",
       elapsed_time(elapsed), (double)num_insn/(double)elapsed);
  if (icache)
    cache_stats(icache, stream);
  if (dcache)
    cache_stats(dcache, stream);
  if (tlb)
    cache_stats(tlb, stream);
}

/* configure the execution engine */
#define SET_NPC(EXPR)		(next_PC = (EXPR))
#define CPC			(regs_PC)

#define GPR(N)			(regs_R[N])
#define GPR_D(N)		(regs_R[N >> 1])
#define SET_GPR(N,EXPR)		(regs_R[N] = (EXPR))

#define FPR_L(N)		(regs_F.l[(N)])
#define SET_FPR_L(N,EXPR)	(regs_F.l[(N)] = (EXPR))
#define FPR_F(N)		(regs_F.f[(N)])
#define SET_FPR_F(N,EXPR)	(regs_F.f[(N)] = (EXPR))
#define FPR_D(N)		(regs_F.d[(N) >> 1])
#define SET_FPR_D(N,EXPR)	(regs_F.d[(N) >> 1] = (EXPR))

#define SET_HI(EXPR)		(regs_HI = (EXPR))
#define HI			(regs_HI)
#define SET_LO(EXPR)		(regs_LO = (EXPR))
#define LO			(regs_LO)
#define FCC			(regs_FCC)
#define SET_FCC(EXPR)		(regs_FCC = (EXPR))

#define __READ_CACHE(DST_T, SRC_T, SRC)					\
  ({ SRC_T temp; num_refs++;						\
     if (tlb)								\
       cache_access(tlb, Read, (SRC), NULL, sizeof(SRC_T), 0, NULL, NULL);	\
     if (dcache)							\
       cache_access(dcache, Read, (SRC), &temp, sizeof(SRC_T), 0, NULL, NULL);\
     else								\
       mem_access(Read, (SRC), &temp, sizeof(SRC_T));			\
     ((unsigned long)((DST_T)temp)); })

#define READ_WORD(SRC)							\
  __READ_CACHE(unsigned long, unsigned long, (SRC))

#define READ_UNSIGNED_HALF(SRC)						\
  __READ_CACHE(unsigned long, unsigned short, (SRC))

#define READ_SIGNED_HALF(SRC)						\
  __READ_CACHE(signed long, signed short, (SRC))

#define READ_UNSIGNED_BYTE(SRC)						\
  __READ_CACHE(unsigned long, unsigned char, (SRC))

#define READ_SIGNED_BYTE(SRC)						\
  __READ_CACHE(signed long, signed char, (SRC))

#define __WRITE_CACHE(SRC, DST_T, DST)					\
  ({ DST_T temp = (DST_T)((unsigned long)(SRC));			\
     num_refs++;							\
     if (tlb)								\
       cache_access(tlb, Write, (DST), NULL, sizeof(DST_T), 0, NULL, NULL);	\
     if (dcache)							\
       cache_access(dcache, Write, (DST), &temp, sizeof(DST_T), 0, NULL, NULL);\
     else								\
       mem_access(Write, (DST), &temp, sizeof(DST_T)); })

#define WRITE_WORD(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned long, (DST))

#define WRITE_HALF(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned short, (DST))

#define WRITE_BYTE(SRC,DST)						\
  __WRITE_CACHE((SRC), unsigned char, (DST))

void
dcache_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE addr, void *p, int nbytes)
{
  if (tlb)
    cache_access(tlb, cmd, addr, NULL, nbytes, 0, NULL, NULL);
  if (dcache)
    cache_access(dcache, cmd, addr, p, nbytes, 0, NULL, NULL);
  else
    mem_access(cmd, addr, p, nbytes);
}

#define SYSCALL(INST)							\
  (flush_on_syscalls							\
   ? ((tlb ? cache_flush(tlb, 0) : 0),					\
      (dcache ? cache_flush(dcache, 0) : 0),				\
      ss_syscall(mem_access, INST))					\
   : (ss_syscall(dcache_access_fn, INST)))

#define DEFFU(FU,DESC)
#define DEFICLASS(ICLASS,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLDST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR,DIRECT)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFFU
#undef DEFICLASS
#undef DEFINST
#undef DEFLDST
#undef DEFLINK
#undef CONNECT
#undef IMPL

/* exit when this becomes non-zero */
int sim_exit_now = FALSE;

/* longjmp here when simulation is completed */
jmp_buf sim_exit_buf;

/* instruction jump table */
static void *op_jump[];

void
sim_main(void)
{
  SS_INST_TYPE inst;
  register SS_ADDR_TYPE next_PC;

  /* instruction jump table */
  static void *op_jump[] = {
    &&opcode_NA, /* NA */
#define DEFFU(FU,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
    &&opcode_##OP,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
    &&opcode_##OP,
#define CONNECT(OP)
#include "ss.def"
#undef DEFFU
#undef DEFINST
#undef DEFLINK
#undef CONNECT
  };

  info("starting functional simulation");

  num_insn = 0;
  num_refs = 0;

  regs_PC = ld_prog_entry - SS_INST_SIZE;
  next_PC = ld_prog_entry;

  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    fprintf(stderr, "decoding text segment...");
    for (addr=ld_text_base;
	 addr < (ld_text_base+ld_text_size);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr);
	inst.a = (inst.a & ~0xff) | (unsigned long)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr) = inst;
      }
    fprintf(stderr, "done.\n");
  }

  if (icache)
    cache_access(icache, Read, next_PC, &inst, sizeof(SS_INST_TYPE), 0, NULL, NULL);
  else
    mem_access(Read, next_PC, &inst, sizeof(SS_INST_TYPE));
  goto *op_jump[SS_OPCODE(inst)];

#define DEFFU(FU,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
  opcode_##OP:								\
    /* maintain $r0 semantics */					\
    regs_R[0] = 0;							\
    /* keep an instruction count */					\
    num_insn++;								\
    /* execute next instruction */					\
    regs_PC = next_PC;							\
    next_PC += 8;							\
    EXPR;								\
    if (icache)								\
      cache_access(icache, Read, next_PC, &inst,			\
		   sizeof(SS_INST_TYPE), 0, NULL, NULL);		\
    else								\
      mem_access(Read, next_PC, &inst, sizeof(SS_INST_TYPE));		\
    goto *op_jump[SS_OPCODE(inst)];
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
  opcode_##OP:								\
    panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#include "ss.def"
#undef DEFFU
#undef DEFINST
#undef DEFLINK
#undef CONNECT
 opcode_NA:
    panic("attempted to execute a bogus opcode");
}
